/*
 * gsmtstr - GSM T(urbo) STR(ing) conversion utility
 *
 * Copyright (C) 2004, 2005 BLADOX, s.r.o.
 *
 * www.bladox.com
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/*
#
#	Name:             GSM 03.38 to Unicode
#	Unicode version:  3.0
#	Table version:    1.1
#	Table format:     Format A
#	Date:             2000 May 30
#	Authors:          Ken Whistler
#                         Kent Karlsson
#                         Markus Kuhn
#
#	Copyright (c) 2000 Unicode, Inc.  All Rights reserved.
#
#	This file is provided as-is by Unicode, Inc. (The Unicode Consortium).
#	No claims are made as to fitness for any particular purpose.  No
#	warranties of any kind are expressed or implied.  The recipient
#	agrees to determine applicability of information provided.  If this
#	file has been provided on optical media by Unicode, Inc., the sole
#	remedy for any claim will be exchange of defective media within 90
#	days of receipt.
#
#	Unicode, Inc. hereby grants the right to freely use the information
#	supplied in this file in the creation of products supporting the
#	Unicode Standard, and to make copies of this file in any form for
#	internal or external distribution as long as this notice remains
#	attached.
#
#	General notes:
#
#	This table contains the data the Unicode Consortium has on how
#       ETSI GSM 03.38 7-bit default alphabet characters map into Unicode.
#	This mapping is based on ETSI TS 100 900 V7.2.0 (1999-07), with
#	a correction of 0x09 to *small* c-cedilla, instead of *capital*
#	C-cedilla.
#
#	Format:  Three tab-separated columns
#		 Column #1 is the ETSI GSM 03.38 7-bit default alphabet 
#                             code (in hex as 0xXX, or 0xXXXX for double-byte
#                             sequences)
#		 Column #2 is the Unicode scalar value (in hex as 0xXXXX)
#		 Column #3 the Unicode name (follows a comment sign, '#')
#
#	The entries are in ETSI GSM 03.38 7-bit default alphabet code order.
#
#       Note that ETSI GSM 03.38 also allows for the use of UCS-2 (UTF-16
#       restricted to the BMP) in GSM/SMS messages.
#
#	Note also that there are commented Greek mappings for some 
#	capital Latin characters. This follows from the clear intent
#	of the ETSI GSM 03.38 to have glyph coverage for the uppercase
#	Greek alphabet by reusing Latin letters that have the same 
#	form as an uppercase Greek letter. Conversion implementations 
#	should be aware of this fact.
#
#       The ETSI GSM 03.38 specification shows an uppercase C-cedilla
#       glyph at 0x09. This may be the result of limited display
#       capabilities for handling characters with descenders. However, the
#       language coverage intent is clearly for the lowercase c-cedilla, as shown
#       in the mapping below. The mapping for uppercase C-cedilla is shown
#       in a commented line in the mapping table.
#
#	The ESC character 0x1B is
#	mapped to the no-break space character, unless it is part of a
#	valid ESC sequence, to facilitate round-trip compatibility in
#	the presence of unknown ESC sequences.
#
#	0x00 is NULL (when followed only by 0x00 up to the
#	end of (fixed byte length) message, possibly also up to
#	FORM FEED.  But 0x00 is also the code for COMMERCIAL AT
#	when some other character (CARRIAGE RETURN if nothing else)
#	comes after the 0x00.
#
#	Version history
#	1.0 version: first creation
#	1.1 version: fixed problem with the wrong line being a comment,
#			added text regarding 0x00's interpretation,
#                       added second mapping for C-cedilla,
#                       added mapping of 0x1B escape to NBSP for display.
#
#	Updated versions of this file may be found in:
#		<ftp://ftp.unicode.org/Public/MAPPINGS/>
#
#	Any comments or problems, contact <errata@unicode.org>
#	Please note that <errata@unicode.org> is an archival address;
#	notices will be checked, but do not expect an immediate response.
#
0x00	0x0040	#	COMMERCIAL AT
#0x00	0x0000	#	NULL (see note above)
0x01	0x00A3	#	POUND SIGN
0x02	0x0024	#	DOLLAR SIGN
0x03	0x00A5	#	YEN SIGN
0x04	0x00E8	#	LATIN SMALL LETTER E WITH GRAVE
0x05	0x00E9	#	LATIN SMALL LETTER E WITH ACUTE
0x06	0x00F9	#	LATIN SMALL LETTER U WITH GRAVE
0x07	0x00EC	#	LATIN SMALL LETTER I WITH GRAVE
0x08	0x00F2	#	LATIN SMALL LETTER O WITH GRAVE
0x09	0x00E7	#	LATIN SMALL LETTER C WITH CEDILLA
#0x09	0x00C7	#	LATIN CAPITAL LETTER C WITH CEDILLA (see note above)
0x0A	0x000A	#	LINE FEED
0x0B	0x00D8	#	LATIN CAPITAL LETTER O WITH STROKE
0x0C	0x00F8	#	LATIN SMALL LETTER O WITH STROKE
0x0D	0x000D	#	CARRIAGE RETURN
0x0E	0x00C5	#	LATIN CAPITAL LETTER A WITH RING ABOVE
0x0F	0x00E5	#	LATIN SMALL LETTER A WITH RING ABOVE
0x10	0x0394	#	GREEK CAPITAL LETTER DELTA
0x11	0x005F	#	LOW LINE
0x12	0x03A6	#	GREEK CAPITAL LETTER PHI
0x13	0x0393	#	GREEK CAPITAL LETTER GAMMA
0x14	0x039B	#	GREEK CAPITAL LETTER LAMDA
0x15	0x03A9	#	GREEK CAPITAL LETTER OMEGA
0x16	0x03A0	#	GREEK CAPITAL LETTER PI
0x17	0x03A8	#	GREEK CAPITAL LETTER PSI
0x18	0x03A3	#	GREEK CAPITAL LETTER SIGMA
0x19	0x0398	#	GREEK CAPITAL LETTER THETA
0x1A	0x039E	#	GREEK CAPITAL LETTER XI
0x1B	0x00A0	#	ESCAPE TO EXTENSION TABLE (or displayed as NBSP, see note above)
0x1B0A	0x000C	#	FORM FEED
0x1B14	0x005E	#	CIRCUMFLEX ACCENT
0x1B28	0x007B	#	LEFT CURLY BRACKET
0x1B29	0x007D	#	RIGHT CURLY BRACKET
0x1B2F	0x005C	#	REVERSE SOLIDUS
0x1B3C	0x005B	#	LEFT SQUARE BRACKET
0x1B3D	0x007E	#	TILDE
0x1B3E	0x005D	#	RIGHT SQUARE BRACKET
0x1B40	0x007C	#	VERTICAL LINE
0x1B65	0x20AC	#	EURO SIGN
0x1C	0x00C6	#	LATIN CAPITAL LETTER AE
0x1D	0x00E6	#	LATIN SMALL LETTER AE
0x1E	0x00DF	#	LATIN SMALL LETTER SHARP S (German)
0x1F	0x00C9	#	LATIN CAPITAL LETTER E WITH ACUTE
0x20	0x0020	#	SPACE
0x21	0x0021	#	EXCLAMATION MARK
0x22	0x0022	#	QUOTATION MARK
0x23	0x0023	#	NUMBER SIGN
0x24	0x00A4	#	CURRENCY SIGN
0x25	0x0025	#	PERCENT SIGN
0x26	0x0026	#	AMPERSAND
0x27	0x0027	#	APOSTROPHE
0x28	0x0028	#	LEFT PARENTHESIS
0x29	0x0029	#	RIGHT PARENTHESIS
0x2A	0x002A	#	ASTERISK
0x2B	0x002B	#	PLUS SIGN
0x2C	0x002C	#	COMMA
0x2D	0x002D	#	HYPHEN-MINUS
0x2E	0x002E	#	FULL STOP
0x2F	0x002F	#	SOLIDUS
0x30	0x0030	#	DIGIT ZERO
0x31	0x0031	#	DIGIT ONE
0x32	0x0032	#	DIGIT TWO
0x33	0x0033	#	DIGIT THREE
0x34	0x0034	#	DIGIT FOUR
0x35	0x0035	#	DIGIT FIVE
0x36	0x0036	#	DIGIT SIX
0x37	0x0037	#	DIGIT SEVEN
0x38	0x0038	#	DIGIT EIGHT
0x39	0x0039	#	DIGIT NINE
0x3A	0x003A	#	COLON
0x3B	0x003B	#	SEMICOLON
0x3C	0x003C	#	LESS-THAN SIGN
0x3D	0x003D	#	EQUALS SIGN
0x3E	0x003E	#	GREATER-THAN SIGN
0x3F	0x003F	#	QUESTION MARK
0x40	0x00A1	#	INVERTED EXCLAMATION MARK
0x41	0x0041	#	LATIN CAPITAL LETTER A
#0x41	0x0391	#	GREEK CAPITAL LETTER ALPHA
0x42	0x0042	#	LATIN CAPITAL LETTER B
#0x42	0x0392	#	GREEK CAPITAL LETTER BETA
0x43	0x0043	#	LATIN CAPITAL LETTER C
0x44	0x0044	#	LATIN CAPITAL LETTER D
0x45	0x0045	#	LATIN CAPITAL LETTER E
#0x45	0x0395	#	GREEK CAPITAL LETTER EPSILON
0x46	0x0046	#	LATIN CAPITAL LETTER F
0x47	0x0047	#	LATIN CAPITAL LETTER G
0x48	0x0048	#	LATIN CAPITAL LETTER H
#0x48	0x0397	#	GREEK CAPITAL LETTER ETA
0x49	0x0049	#	LATIN CAPITAL LETTER I
#0x49	0x0399	#	GREEK CAPITAL LETTER IOTA
0x4A	0x004A	#	LATIN CAPITAL LETTER J
0x4B	0x004B	#	LATIN CAPITAL LETTER K
#0x4B	0x039A	#	GREEK CAPITAL LETTER KAPPA
0x4C	0x004C	#	LATIN CAPITAL LETTER L
0x4D	0x004D	#	LATIN CAPITAL LETTER M
#0x4D	0x039C	#	GREEK CAPITAL LETTER MU
0x4E	0x004E	#	LATIN CAPITAL LETTER N
#0x4E	0x039D	#	GREEK CAPITAL LETTER NU
0x4F	0x004F	#	LATIN CAPITAL LETTER O
#0x4F	0x039F	#	GREEK CAPITAL LETTER OMICRON
0x50	0x0050	#	LATIN CAPITAL LETTER P
#0x50	0x03A1	#	GREEK CAPITAL LETTER RHO
0x51	0x0051	#	LATIN CAPITAL LETTER Q
0x52	0x0052	#	LATIN CAPITAL LETTER R
0x53	0x0053	#	LATIN CAPITAL LETTER S
0x54	0x0054	#	LATIN CAPITAL LETTER T
#0x54	0x03A4	#	GREEK CAPITAL LETTER TAU
0x55	0x0055	#	LATIN CAPITAL LETTER U
#0x55	0x03A5	#	GREEK CAPITAL LETTER UPSILON
0x56	0x0056	#	LATIN CAPITAL LETTER V
0x57	0x0057	#	LATIN CAPITAL LETTER W
0x58	0x0058	#	LATIN CAPITAL LETTER X
#0x58	0x03A7	#	GREEK CAPITAL LETTER CHI
0x59	0x0059	#	LATIN CAPITAL LETTER Y
0x5A	0x005A	#	LATIN CAPITAL LETTER Z
#0x5A	0x0396	#	GREEK CAPITAL LETTER ZETA
0x5B	0x00C4	#	LATIN CAPITAL LETTER A WITH DIAERESIS
0x5C	0x00D6	#	LATIN CAPITAL LETTER O WITH DIAERESIS
0x5D	0x00D1	#	LATIN CAPITAL LETTER N WITH TILDE
0x5E	0x00DC	#	LATIN CAPITAL LETTER U WITH DIAERESIS
0x5F	0x00A7	#	SECTION SIGN
0x60	0x00BF	#	INVERTED QUESTION MARK
0x61	0x0061	#	LATIN SMALL LETTER A
0x62	0x0062	#	LATIN SMALL LETTER B
0x63	0x0063	#	LATIN SMALL LETTER C
0x64	0x0064	#	LATIN SMALL LETTER D
0x65	0x0065	#	LATIN SMALL LETTER E
0x66	0x0066	#	LATIN SMALL LETTER F
0x67	0x0067	#	LATIN SMALL LETTER G
0x68	0x0068	#	LATIN SMALL LETTER H
0x69	0x0069	#	LATIN SMALL LETTER I
0x6A	0x006A	#	LATIN SMALL LETTER J
0x6B	0x006B	#	LATIN SMALL LETTER K
0x6C	0x006C	#	LATIN SMALL LETTER L
0x6D	0x006D	#	LATIN SMALL LETTER M
0x6E	0x006E	#	LATIN SMALL LETTER N
0x6F	0x006F	#	LATIN SMALL LETTER O
0x70	0x0070	#	LATIN SMALL LETTER P
0x71	0x0071	#	LATIN SMALL LETTER Q
0x72	0x0072	#	LATIN SMALL LETTER R
0x73	0x0073	#	LATIN SMALL LETTER S
0x74	0x0074	#	LATIN SMALL LETTER T
0x75	0x0075	#	LATIN SMALL LETTER U
0x76	0x0076	#	LATIN SMALL LETTER V
0x77	0x0077	#	LATIN SMALL LETTER W
0x78	0x0078	#	LATIN SMALL LETTER X
0x79	0x0079	#	LATIN SMALL LETTER Y
0x7A	0x007A	#	LATIN SMALL LETTER Z
0x7B	0x00E4	#	LATIN SMALL LETTER A WITH DIAERESIS
0x7C	0x00F6	#	LATIN SMALL LETTER O WITH DIAERESIS
0x7D	0x00F1	#	LATIN SMALL LETTER N WITH TILDE
0x7E	0x00FC	#	LATIN SMALL LETTER U WITH DIAERESIS
0x7F	0x00E0	#	LATIN SMALL LETTER A WITH GRAVE
*/

typedef struct _G7_uni
{
  unsigned short b7;
  unsigned short uni;
} G7_uni;

G7_uni conv_tab_g7_iso_8859_1[] = {
  {0x00, 0x0040},
//{     #0x00, 0x0000},
  {0x01, 0x00A3},
  {0x02, 0x0024},
  {0x03, 0x00A5},
  {0x04, 0x00E8},
  {0x05, 0x00E9},
  {0x06, 0x00F9},
  {0x07, 0x00EC},
  {0x08, 0x00F2},
  {0x09, 0x00E7},
//{     #0x09, 0x00C7},
  {0x0A, 0x000A},
  {0x0B, 0x00D8},
  {0x0C, 0x00F8},
  {0x0D, 0x000D},
  {0x0E, 0x00C5},
  {0x0F, 0x00E5},
  {0x10, 0x0394},
  {0x11, 0x005F},
  {0x12, 0x03A6},
  {0x13, 0x0393},
  {0x14, 0x039B},
  {0x15, 0x03A9},
  {0x16, 0x03A0},
  {0x17, 0x03A8},
  {0x18, 0x03A3},
  {0x19, 0x0398},
  {0x1A, 0x039E},
  {0x1B, 0x00A0},
  {0x1B0A, 0x000C},
  {0x1B14, 0x005E},
  {0x1B28, 0x007B},
  {0x1B29, 0x007D},
  {0x1B2F, 0x005C},
  {0x1B3C, 0x005B},
  {0x1B3D, 0x007E},
  {0x1B3E, 0x005D},
  {0x1B40, 0x007C},
  {0x1B65, 0x20AC},
  {0x1C, 0x00C6},
  {0x1D, 0x00E6},
  {0x1E, 0x00DF},
  {0x1F, 0x00C9},
  {0x20, 0x0020},
  {0x21, 0x0021},
  {0x22, 0x0022},
  {0x23, 0x0023},
  {0x24, 0x00A4},
  {0x25, 0x0025},
  {0x26, 0x0026},
  {0x27, 0x0027},
  {0x28, 0x0028},
  {0x29, 0x0029},
  {0x2A, 0x002A},
  {0x2B, 0x002B},
  {0x2C, 0x002C},
  {0x2D, 0x002D},
  {0x2E, 0x002E},
  {0x2F, 0x002F},
  {0x30, 0x0030},
  {0x31, 0x0031},
  {0x32, 0x0032},
  {0x33, 0x0033},
  {0x34, 0x0034},
  {0x35, 0x0035},
  {0x36, 0x0036},
  {0x37, 0x0037},
  {0x38, 0x0038},
  {0x39, 0x0039},
  {0x3A, 0x003A},
  {0x3B, 0x003B},
  {0x3C, 0x003C},
  {0x3D, 0x003D},
  {0x3E, 0x003E},
  {0x3F, 0x003F},
  {0x40, 0x00A1},
  {0x41, 0x0041},
//{     #0x41, 0x0391},
  {0x42, 0x0042},
//{     #0x42, 0x0392},
  {0x43, 0x0043},
  {0x44, 0x0044},
  {0x45, 0x0045},
//{     #0x45, 0x0395},
  {0x46, 0x0046},
  {0x47, 0x0047},
  {0x48, 0x0048},
//{     #0x48, 0x0397},
  {0x49, 0x0049},
//{     #0x49, 0x0399},
  {0x4A, 0x004A},
  {0x4B, 0x004B},
//{     #0x4B, 0x039A},
  {0x4C, 0x004C},
  {0x4D, 0x004D},
//{     #0x4D, 0x039C},
  {0x4E, 0x004E},
//{     #0x4E, 0x039D},
  {0x4F, 0x004F},
//{     #0x4F, 0x039F},
  {0x50, 0x0050},
//{     #0x50, 0x03A1},
  {0x51, 0x0051},
  {0x52, 0x0052},
  {0x53, 0x0053},
  {0x54, 0x0054},
//{     #0x54, 0x03A4},
  {0x55, 0x0055},
//{     #0x55, 0x03A5},
  {0x56, 0x0056},
  {0x57, 0x0057},
  {0x58, 0x0058},
//{     #0x58, 0x03A7},
  {0x59, 0x0059},
  {0x5A, 0x005A},
//{     #0x5A, 0x0396},
  {0x5B, 0x00C4},
  {0x5C, 0x00D6},
  {0x5D, 0x00D1},
  {0x5E, 0x00DC},
  {0x5F, 0x00A7},
  {0x60, 0x00BF},
  {0x61, 0x0061},
  {0x62, 0x0062},
  {0x63, 0x0063},
  {0x64, 0x0064},
  {0x65, 0x0065},
  {0x66, 0x0066},
  {0x67, 0x0067},
  {0x68, 0x0068},
  {0x69, 0x0069},
  {0x6A, 0x006A},
  {0x6B, 0x006B},
  {0x6C, 0x006C},
  {0x6D, 0x006D},
  {0x6E, 0x006E},
  {0x6F, 0x006F},
  {0x70, 0x0070},
  {0x71, 0x0071},
  {0x72, 0x0072},
  {0x73, 0x0073},
  {0x74, 0x0074},
  {0x75, 0x0075},
  {0x76, 0x0076},
  {0x77, 0x0077},
  {0x78, 0x0078},
  {0x79, 0x0079},
  {0x7A, 0x007A},
  {0x7B, 0x00E4},
  {0x7C, 0x00F6},
  {0x7D, 0x00F1},
  {0x7E, 0x00FC},
  {0x7F, 0x00E0},

// HU substitutions
  {'U', 0x00DA},		// LATIN CAPITAL LETTER U WITH ACUTE
  {'a', 0x00E1},		// LATIN SMALL LETTER A WITH ACUTE
  {'i', 0x00ED},		// LATIN SMALL LETTER I WITH ACUTE
  {'o', 0x00F3},		// LATIN SMALL LETTER O WITH ACUTE
  {0x5C, 0x00F5},		// LATIN SMALL LETTER O WITH DOUBLE ACUTE
  {'u', 0x00FA},		// LATIN SMALL LETTER U WITH ACUTE
  {0x5E, 0x00FB},		// LATIN CAPITAL LETTER U WITH DIAERESIS
};

G7_uni conv_tab_g7_iso_8859_7[] = {	// GR ISO 8859-7 -> 7b
  {'W', 0xB4},			//GREEK TONOS
  {'W', 0xB5},			//GREEK DIALYTIKA TONOS
  {'W', 0xB6},			//GREEK CAPITAL LETTER ALPHA WITH TONOS
  {'W', 0xB8},			//GREEK CAPITAL LETTER EPSILON WITH TONOS
  {'W', 0xB9},			//GREEK CAPITAL LETTER ETA WITH TONOS
  {'W', 0xBA},			//GREEK CAPITAL LETTER IOTA WITH TONOS
  {'W', 0xBC},			//GREEK CAPITAL LETTER OMICRON WITH TONOS
  {'W', 0xBE},			//GREEK CAPITAL LETTER UPSILON WITH TONOS
  {'W', 0xBF},			//GREEK CAPITAL LETTER OMEGA WITH TONOS
  {'W', 0xC0},			//GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS

  {'A', 0xC1},			//GREEK CAPITAL LETTER ALPHA
  {'B', 0xC2},			//GREEK CAPITAL LETTER BETA
  {0x13, 0xC3},			//GREEK CAPITAL LETTER GAMMA
  {0x10, 0xC4},			//GREEK CAPITAL LETTER DELTA
  {'E', 0xC5},			//GREEK CAPITAL LETTER EPSILON
  {'Z', 0xC6},			//GREEK CAPITAL LETTER ZETA
  {'H', 0xC7},			//GREEK CAPITAL LETTER ETA
  {0x19, 0xC8},			//GREEK CAPITAL LETTER THETA
  {'I', 0xC9},			//GREEK CAPITAL LETTER IOTA
  {'K', 0xCA},			//GREEK CAPITAL LETTER KAPPA
  {0x14, 0xCB},			//GREEK CAPITAL LETTER LAMDA
  {'M', 0xCC},			//GREEK CAPITAL LETTER MU
  {'N', 0xCD},			//GREEK CAPITAL LETTER NU
  {0x1A, 0xCE},			//GREEK CAPITAL LETTER XI
  {'O', 0xCF},			//GREEK CAPITAL LETTER OMICRON
  {0x16, 0xD0},			//GREEK CAPITAL LETTER PI
  {'P', 0xD1},			//GREEK CAPITAL LETTER RHO
  {0x18, 0xD3},			//GREEK CAPITAL LETTER SIGMA
  {'T', 0xD4},			//GREEK CAPITAL LETTER TAU
  {'Y', 0xD5},			//GREEK CAPITAL LETTER UPSILON
  {0x12, 0xD6},			//GREEK CAPITAL LETTER PHI
  {'X', 0xD7},			//GREEK CAPITAL LETTER CHI
  {0x17, 0xD8},			//GREEK CAPITAL LETTER PSI
  {0x15, 0xD9},			//GREEK CAPITAL LETTER OMEGA

  {'w', 0xDA},			//GREEK CAPITAL LETTER IOTA WITH DIALYTIKA
  {'w', 0xDB},			//GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
  {'w', 0xDC},			//GREEK SMALL LETTER ALPHA WITH TONOS
  {'w', 0xDD},			//GREEK SMALL LETTER EPSILON WITH TONOS
  {'w', 0xDE},			//GREEK SMALL LETTER ETA WITH TONOS
  {'w', 0xDF},			//GREEK SMALL LETTER IOTA WITH TONOS
  {'w', 0xE0},			//GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS

  {'w', 0xE1},			//GREEK SMALL LETTER ALPHA
  {'w', 0xE2},			//GREEK SMALL LETTER BETA
  {'w', 0xE3},			//GREEK SMALL LETTER GAMMA
  {'w', 0xE4},			//GREEK SMALL LETTER DELTA
  {'w', 0xE5},			//GREEK SMALL LETTER EPSILON
  {'w', 0xE6},			//GREEK SMALL LETTER ZETA
  {'w', 0xE7},			//GREEK SMALL LETTER ETA
  {'w', 0xE8},			//GREEK SMALL LETTER THETA
  {'w', 0xE9},			//GREEK SMALL LETTER IOTA
  {'w', 0xEA},			//GREEK SMALL LETTER KAPPA
  {'w', 0xEB},			//GREEK SMALL LETTER LAMDA
  {'w', 0xEC},			//GREEK SMALL LETTER MU
  {'w', 0xED},			//GREEK SMALL LETTER NU
  {'w', 0xEE},			//GREEK SMALL LETTER XI
  {'w', 0xEF},			//GREEK SMALL LETTER OMICRON
  {'w', 0xF0},			//GREEK SMALL LETTER PI
  {'w', 0xF1},			//GREEK SMALL LETTER RHO
  {'w', 0xF2},			//GREEK SMALL LETTER FINAL SIGMA
  {'w', 0xF3},			//GREEK SMALL LETTER SIGMA
  {'w', 0xF4},			//GREEK SMALL LETTER TAU
  {'w', 0xF5},			//GREEK SMALL LETTER UPSILON
  {'w', 0xF6},			//GREEK SMALL LETTER PHI
  {'w', 0xF7},			//GREEK SMALL LETTER CHI
  {'w', 0xF8},			//GREEK SMALL LETTER PSI
  {'w', 0xF9},			//GREEK SMALL LETTER OMEGA

  {'w', 0xFA},			//GREEK SMALL LETTER IOTA WITH DIALYTIKA
  {'w', 0xFB},			//GREEK SMALL LETTER UPSILON WITH DIALYTIKA
  {'w', 0xFC},			//GREEK SMALL LETTER OMICRON WITH TONOS
  {'w', 0xFD},			//GREEK SMALL LETTER UPSILON WITH TONOS
  {'w', 0xFE},			//GREEK SMALL LETTER OMEGA WITH TONOS

  {0x00, 0x0040},
//{     #0x00, 0x0000},
  {0x01, 0x00A3},
  {0x02, 0x0024},
  {0x03, 0x00A5},

//NNN  {0x04, 0x00E8}, //XXX GREEK SMALL LETTER THETA
//NNN  {0x05, 0x00E9}, //XXX GREEK SMALL LETTER IOTA
//NNN  {0x06, 0x00F9}, //XXX GREEK SMALL LETTER OMEGA
//NNN  {0x07, 0x00EC}, //XXX GREEK SMALL LETTER MU
//NNN  {0x08, 0x00F2}, //XXX GREEK SMALL LETTER FINAL SIGMA
//NNN  {0x09, 0x00E7}, //XXX GREEK SMALL LETTER ETA
//NNN  {0x09, 0x00C7}, //XXX GREEK CAPITAL LETTER ETA
  {0x0A, 0x000A},
//NNN  {0x0B, 0x00D8}, //XXX GREEK CAPITAL LETTER PSI
//NNN  {0x0C, 0x00F8}, //XXX GREEK SMALL LETTER PSI
  {0x0D, 0x000D},
//NNN  {0x0E, 0x00C5}, //XXX GREEK CAPITAL LETTER EPSILON
//NNN  {0x0F, 0x00E5}, //XXX GREEK SMALL LETTER EPSILON
//NNN  {0x10, 0x00C4}, //YYY GREEK CAPITAL LETTER DELTA
  {0x11, 0x005F},
//NNN  {0x12, 0x00D6}, //YYY GREEK CAPITAL LETTER PHI
//NNN  {0x13, 0x00C3}, //YYY GREEK CAPITAL LETTER GAMMA
//NNN  {0x14, 0x00CB}, //YYY GREEK CAPITAL LETTER LAMDA
//NNN  {0x15, 0x00D9}, //YYY GREEK CAPITAL LETTER OMEGA
//NNN  {0x16, 0x00D0}, //YYY GREEK CAPITAL LETTER PI
//NNN  {0x17, 0x00D8}, //YYY GREEK CAPITAL LETTER PSI
//NNN  {0x18, 0x00D3}, //YYY GREEK CAPITAL LETTER SIGMA
//NNN  {0x19, 0x00C8}, //YYY GREEK CAPITAL LETTER THETA
//NNN  {0x1A, 0x00CE}, //YYY GREEK CAPITAL LETTER XI
  {0x1B, 0x00A0},
  {0x1B0A, 0x000C},
  {0x1B14, 0x005E},
  {0x1B28, 0x007B},
  {0x1B29, 0x007D},
  {0x1B2F, 0x005C},
  {0x1B3C, 0x005B},
  {0x1B3D, 0x007E},
  {0x1B3E, 0x005D},
  {0x1B40, 0x007C},
  {0x1B65, 0x20AC},
//NNN  {0x1C, 0x00C6}, //XXX GREEK CAPITAL LETTER ZETA
//NNN  {0x1D, 0x00E6}, //XXX GREEK SMALL LETTER ZETA
//NNN  {0x1E, 0x00DF}, //XXX GREEK SMALL LETTER IOTA WITH TONOS
//NNN  {0x1F, 0x00C9}, //XXX GREEK CAPITAL LETTER IOTA
  {0x20, 0x0020},
  {0x21, 0x0021},
  {0x22, 0x0022},
  {0x23, 0x0023},
  {0x24, 0x00A4},
  {0x25, 0x0025},
  {0x26, 0x0026},
  {0x27, 0x0027},
  {0x28, 0x0028},
  {0x29, 0x0029},
  {0x2A, 0x002A},
  {0x2B, 0x002B},
  {0x2C, 0x002C},
  {0x2D, 0x002D},
  {0x2E, 0x002E},
  {0x2F, 0x002F},
  {0x30, 0x0030},
  {0x31, 0x0031},
  {0x32, 0x0032},
  {0x33, 0x0033},
  {0x34, 0x0034},
  {0x35, 0x0035},
  {0x36, 0x0036},
  {0x37, 0x0037},
  {0x38, 0x0038},
  {0x39, 0x0039},
  {0x3A, 0x003A},
  {0x3B, 0x003B},
  {0x3C, 0x003C},
  {0x3D, 0x003D},
  {0x3E, 0x003E},
  {0x3F, 0x003F},
  {0x40, 0x00A1},
  {0x41, 0x0041},
//NNN  {0x41, 0x0391}, //YYY GREEK CAPITAL LETTER ALPHA
  {0x42, 0x0042},
//NNN  {0x42, 0x0392}, //YYY GREEK CAPITAL LETTER BETA
  {0x43, 0x0043},
  {0x44, 0x0044},
  {0x45, 0x0045},
//NNN  {0x45, 0x0395}, //YYY GREEK CAPITAL LETTER EPSILON
  {0x46, 0x0046},
  {0x47, 0x0047},
  {0x48, 0x0048},
//NNN  {0x48, 0x0397}, //YYY GREEK CAPITAL LETTER ETA
  {0x49, 0x0049},
//NNN  {0x49, 0x0399}, //YYY GREEK CAPITAL LETTER IOTA
  {0x4A, 0x004A},
  {0x4B, 0x004B},
//NNN  {0x4B, 0x039A}, //YYY GREEK CAPITAL LETTER KAPPA
  {0x4C, 0x004C},
  {0x4D, 0x004D},
//NNN  {0x4D, 0x039C}, //YYY GREEK CAPITAL LETTER MU
  {0x4E, 0x004E},
//NNN  {0x4E, 0x039D}, //YYY GREEK CAPITAL LETTER NU
  {0x4F, 0x004F},
//NNN  {0x4F, 0x039F}, //YYY GREEK CAPITAL LETTER OMICRON
  {0x50, 0x0050},
//NNN  {0x50, 0x03A1}, //YYY GREEK CAPITAL LETTER RHO
  {0x51, 0x0051},
  {0x52, 0x0052},
  {0x53, 0x0053},
  {0x54, 0x0054},
//NNN  {0x54, 0x03A4}, //YYY GREEK CAPITAL LETTER TAU
  {0x55, 0x0055},
//NNN  {0x55, 0x03A5}, //YYY GREEK CAPITAL LETTER UPSILON
  {0x56, 0x0056},
  {0x57, 0x0057},
  {0x58, 0x0058},
//NNN  {0x58, 0x03A7}, //YYY GREEK CAPITAL LETTER CHI
  {0x59, 0x0059},
  {0x5A, 0x005A},
//NNN  {0x5A, 0x0396}, //YYY GREEK CAPITAL LETTER ZETA
//NNN  {0x5B, 0x00C4}, //XXX GREEK CAPITAL LETTER DELTA
//NNN  {0x5C, 0x00D6}, //XXX GREEK CAPITAL LETTER PHI
//NNN  {0x5D, 0x00D1}, //XXX GREEK CAPITAL LETTER RHO
//NNN  {0x5E, 0x00DC}, //XXX GREEK SMALL LETTER ALPHA WITH TONOS
  {0x5F, 0x00A7},
//NNN  {0x60, 0x00BF}, //XXX GREEK CAPITAL LETTER OMEGA WITH TONOS
  {0x61, 0x0061},
  {0x62, 0x0062},
  {0x63, 0x0063},
  {0x64, 0x0064},
  {0x65, 0x0065},
  {0x66, 0x0066},
  {0x67, 0x0067},
  {0x68, 0x0068},
  {0x69, 0x0069},
  {0x6A, 0x006A},
  {0x6B, 0x006B},
  {0x6C, 0x006C},
  {0x6D, 0x006D},
  {0x6E, 0x006E},
  {0x6F, 0x006F},
  {0x70, 0x0070},
  {0x71, 0x0071},
  {0x72, 0x0072},
  {0x73, 0x0073},
  {0x74, 0x0074},
  {0x75, 0x0075},
  {0x76, 0x0076},
  {0x77, 0x0077},
  {0x78, 0x0078},
  {0x79, 0x0079},
  {0x7A, 0x007A},
//NNN  {0x7B, 0x00E4}, //XXX GREEK SMALL LETTER DELTA
//NNN  {0x7C, 0x00F6}, //XXX GREEK SMALL LETTER PHI
//NNN  {0x7D, 0x00F1}, //XXX GREEK SMALL LETTER RHO
//NNN  {0x7E, 0x00FC}, //XXX GREEK SMALL LETTER OMICRON WITH TONOS
//NNN  {0x7F, 0x00E0}, //XXX GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS
};

G7_uni conv_tab_g7_uni[] = {	// GR chars->7b
  {0x00, 0x0040},
//{     #0x00, 0x0000},
  {0x01, 0x00A3},
  {0x02, 0x0024},
  {0x03, 0x00A5},
  {0x04, 0x00E8},
  {0x05, 0x00E9},
  {0x06, 0x00F9},
  {0x07, 0x00EC},
  {0x08, 0x00F2},
  {0x09, 0x00E7},
  {0x09, 0x00C7},
  {0x0A, 0x000A},
  {0x0B, 0x00D8},
  {0x0C, 0x00F8},
  {0x0D, 0x000D},
  {0x0E, 0x00C5},
  {0x0F, 0x00E5},
  {0x10, 0x0394},
  {0x11, 0x005F},
  {0x12, 0x03A6},
  {0x13, 0x0393},
  {0x14, 0x039B},
  {0x15, 0x03A9},
  {0x16, 0x03A0},
  {0x17, 0x03A8},
  {0x18, 0x03A3},
  {0x19, 0x0398},
  {0x1A, 0x039E},
  {0x1B, 0x00A0},
  {0x1B0A, 0x000C},
  {0x1B14, 0x005E},
  {0x1B28, 0x007B},
  {0x1B29, 0x007D},
  {0x1B2F, 0x005C},
  {0x1B3C, 0x005B},
  {0x1B3D, 0x007E},
  {0x1B3E, 0x005D},
  {0x1B40, 0x007C},
  {0x1B65, 0x20AC},
  {0x1C, 0x00C6},
  {0x1D, 0x00E6},
  {0x1E, 0x00DF},
  {0x1F, 0x00C9},
  {0x20, 0x0020},
  {0x21, 0x0021},
  {0x22, 0x0022},
  {0x23, 0x0023},
  {0x24, 0x00A4},
  {0x25, 0x0025},
  {0x26, 0x0026},
  {0x27, 0x0027},
  {0x28, 0x0028},
  {0x29, 0x0029},
  {0x2A, 0x002A},
  {0x2B, 0x002B},
  {0x2C, 0x002C},
  {0x2D, 0x002D},
  {0x2E, 0x002E},
  {0x2F, 0x002F},
  {0x30, 0x0030},
  {0x31, 0x0031},
  {0x32, 0x0032},
  {0x33, 0x0033},
  {0x34, 0x0034},
  {0x35, 0x0035},
  {0x36, 0x0036},
  {0x37, 0x0037},
  {0x38, 0x0038},
  {0x39, 0x0039},
  {0x3A, 0x003A},
  {0x3B, 0x003B},
  {0x3C, 0x003C},
  {0x3D, 0x003D},
  {0x3E, 0x003E},
  {0x3F, 0x003F},
  {0x40, 0x00A1},
  {0x41, 0x0041},
  {0x41, 0x0391},
  {0x42, 0x0042},
  {0x42, 0x0392},
  {0x43, 0x0043},
  {0x44, 0x0044},
  {0x45, 0x0045},
  {0x45, 0x0395},
  {0x46, 0x0046},
  {0x47, 0x0047},
  {0x48, 0x0048},
  {0x48, 0x0397},
  {0x49, 0x0049},
  {0x49, 0x0399},
  {0x4A, 0x004A},
  {0x4B, 0x004B},
  {0x4B, 0x039A},
  {0x4C, 0x004C},
  {0x4D, 0x004D},
  {0x4D, 0x039C},
  {0x4E, 0x004E},
  {0x4E, 0x039D},
  {0x4F, 0x004F},
  {0x4F, 0x039F},
  {0x50, 0x0050},
  {0x50, 0x03A1},
  {0x51, 0x0051},
  {0x52, 0x0052},
  {0x53, 0x0053},
  {0x54, 0x0054},
  {0x54, 0x03A4},
  {0x55, 0x0055},
  {0x55, 0x03A5},
  {0x56, 0x0056},
  {0x57, 0x0057},
  {0x58, 0x0058},
  {0x58, 0x03A7},
  {0x59, 0x0059},
  {0x5A, 0x005A},
  {0x5A, 0x0396},
  {0x5B, 0x00C4},
  {0x5C, 0x00D6},
  {0x5D, 0x00D1},
  {0x5E, 0x00DC},
  {0x5F, 0x00A7},
  {0x60, 0x00BF},
  {0x61, 0x0061},
  {0x62, 0x0062},
  {0x63, 0x0063},
  {0x64, 0x0064},
  {0x65, 0x0065},
  {0x66, 0x0066},
  {0x67, 0x0067},
  {0x68, 0x0068},
  {0x69, 0x0069},
  {0x6A, 0x006A},
  {0x6B, 0x006B},
  {0x6C, 0x006C},
  {0x6D, 0x006D},
  {0x6E, 0x006E},
  {0x6F, 0x006F},
  {0x70, 0x0070},
  {0x71, 0x0071},
  {0x72, 0x0072},
  {0x73, 0x0073},
  {0x74, 0x0074},
  {0x75, 0x0075},
  {0x76, 0x0076},
  {0x77, 0x0077},
  {0x78, 0x0078},
  {0x79, 0x0079},
  {0x7A, 0x007A},
  {0x7B, 0x00E4},
  {0x7C, 0x00F6},
  {0x7D, 0x00F1},
  {0x7E, 0x00FC},
  {0x7F, 0x00E0},
};

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

/*
TODO
-7 packed output {...} or STRING_SLV
iso-8859-7 'W'/'w' - not implemented, tonos, dialika, small

output ALPHA UCS2 coding version 1/2/3
*/

void help ()
{
  printf ("Usage: gsmtstr [args] infile\n");
  printf ("-8\t\tinput is UTF-8 (i.e. 2 bytes per char)\n");
  printf ("-i iso \t\tinput ISO-8859-i\n");
  printf ("-l\t\tshow len of string, in chars and bytes\n");
  printf ("-m mangle\te.g. LC_XX(...)\n");
  printf ("-r\t\tinclude original prefixed //\n");
  printf ("-s lines\tskip lines, default 0, includes all skipped if -r\n");
  printf
    ("-w warnlen\twarn if len in chars higher than wanrlen, default 14\n");
  printf ("-h\t\thelp\n");
  printf ("-v\t\tverbose\n");
  exit (-1);
}

int main (int argc, char **argv)
{
  FILE *in;
  int c;
  unsigned char ibuf[1000];
  unsigned char obuf[1000];
  int ilen = 0;
  int olen = 0;
  int i, j;
  unsigned char *filename;
  unsigned verbose = 0;
  unsigned incl_orig = 0;
  unsigned char *mangle = NULL;
  unsigned char was = 0;
  int result;
  unsigned char slash = 0;
  unsigned skip = 0;
  unsigned skipped = 0;
  unsigned first_char = 0;
  unsigned char utf8 = 0;
  unsigned conv_tab_len = sizeof (conv_tab_g7_iso_8859_1);
  G7_uni *conv_tab = &conv_tab_g7_iso_8859_1;
  int iso = 1;
  unsigned int warnlen = 14;
  unsigned char showlen = 0;
  unsigned int chlen, blen;

  do
  {
    result = getopt (argc, argv, "vm:hri:s:8lw:");
    switch (result)
    {
      case 'i':
	iso = atoi (optarg);
	break;
      case 's':
	skip = atoi (optarg);
	break;
      case 'w':
	warnlen = atoi (optarg);
	break;
      case 'l':
	showlen = 1;
	break;
      case 'm':
	mangle = optarg;
	break;
      case '8':
	utf8 = 1;
	break;
      case 'r':
	incl_orig = 1;
	break;
      case 'v':
	verbose++;
	break;
      case 'h':
	help ();
	break;
    }
  }
  while (result > 0);

  if (argc - optind != 1)
    help ();

  if (iso != 1)
  {
    if (iso == 7)
    {
      conv_tab = &conv_tab_g7_iso_8859_7;
      conv_tab_len = sizeof (conv_tab_g7_iso_8859_7);
    }
    else
    {
      fprintf (stderr, "Wrong ISO -i %d\n", iso);
      exit (-1);
    }
  }

  filename = argv[optind++];
  in = fopen (filename, "r");
  if (in == NULL)
  {
    fprintf (stderr, "Cannot open %s\n", filename);
    exit (-1);
  }

  first_char = 1;

  while ((c = fgetc (in)) != EOF)
  {
    slash = 0;
    if (c != '\r')
      if (skipped < skip)
      {
	if (c != '\n' && c != '\r')
	{
	  if (incl_orig)
	  {
	    if (first_char)
	      printf ("// \"");

	    printf ("%c", c);
	  }
	  first_char = 0;
	}
	else
	{
	  if (incl_orig)
	    printf ("\"\n");
	  skipped++;
	  first_char = 1;
	}
      }
      else
      {
	if (c != '\n' && c != '\r')
	{
	  was = 1;

	  if (utf8 == 0)
	  {
	    if (c == '\\')
	    {
	      slash = 1;
	      c = fgetc (in);
	    }

	    for (i = 0; i < conv_tab_len / sizeof (G7_uni); i++)
	    {
	      if ((unsigned char) c == conv_tab[i].uni)
		break;
	    }
	  }

//if(c==0xD5) printf("UPSILON!\n");

	  if (utf8 == 0 && i == conv_tab_len / sizeof (G7_uni))
	  {
	    printf ("ERROR - unknown char: %02X %c\n", c, c);
	    obuf[olen++] = '.';
	    if (slash)
	      ibuf[ilen++] = '\\';
	    ibuf[ilen++] = c;
	  }
	  else
	  {
	    if (utf8 == 0 && conv_tab[i].uni != conv_tab[i].b7)
	    {
	      if (verbose)
		printf ("Conv: %02X %c -> %02X %c\n", c, c,
			conv_tab[i].b7, conv_tab[i].b7);
	      if (slash)
		ibuf[ilen++] = '\\';
	      ibuf[ilen++] = c;
	      if (slash)
		obuf[olen++] = '\\';
	      if (conv_tab[i].b7 & 0xFF00)
		obuf[olen++] = (conv_tab[i].b7 >> 8) | (0x80);
	      obuf[olen++] = conv_tab[i].b7 | (0x80);
	    }
	    else
	    {
	      if (verbose)
		printf ("NoConv: %02X %c\n", c, c);
	      if (slash)
		obuf[olen++] = '\\';
	      obuf[olen++] = c;
	      if (slash)
		ibuf[ilen++] = '\\';
	      ibuf[ilen++] = c;
	    }
	  }
	}
	else
	{
	  skipped = 0;
	  if (was == 1)
	  {
	    if (incl_orig)
	    {
	      printf ("// \"");
	      for (i = 0; i < ilen; i++)
		printf ("%c", ibuf[i]);
	      printf ("\"\n");
	    }

	    i = 0;
	    chlen = 0;
	    blen = 0;

	    if (utf8)
	      while (i < olen)
	      {
		unsigned short xx = obuf[i];

		if ((xx & 0x80) == 0)
		{
// 1 byte ascii
		  blen += 2;
		}
		else
		{
// 2 bytes utf-8
		  i++;
		  blen += 2;
		}

		i++;
		chlen++;
	      }
	    else
	    {
	      chlen = blen = olen;
	    }

	    if (showlen)
	    {
	      printf ("// Len %d chars, %d bytes\n", chlen, blen);
	      if (chlen > warnlen)
		printf
		  ("// WARNING: len %d chars, %d bytes is higher than %d chars\n",
		   chlen, blen, warnlen);
	    }

	    if (mangle)
	      printf ("%s(", mangle);

	    printf ("\"");

#define STRING_SLV      0x84

	    if (utf8)
	    {
// XXX/TODO ALPHA vs. DCS strings in kernel
/*
STRING_SLV, DCS (=0x08), LEN in bytes, 0x80 - UCS2v1
STRING_SLV, DCS (=0x08), LEN in bytes, 0x81 - UCS2v2, #of chars, page
STRING_SLV, DCS (=0x08), LEN in bytes, 0x82 - UCS2v3, #of chars, page_H, page_L
*/
	      blen += 1;	//0x80
	      printf ("\\%o\\%o\\%o\\%o", STRING_SLV, 0x08, blen, 0x80);

	      i = 0;
	      while (i < olen)
	      {
		unsigned short xx = obuf[i];

		if ((xx & 0x80) == 0)
		{
		  // 1 byte ascii
		  printf ("\\%o\\%o", 0x00, xx);
		}
		else
		{
		  // 2 bytes utf-8
		  xx &= 0x3F;
		  xx <<= 6;
		  i++;
		  xx = xx | (obuf[i] & 0x3F);
		  printf ("\\%o\\%o", (xx >> 8), xx & 0xFF);
		}

		i++;
	      }
	    }
	    else
	      for (i = 0; i < olen; i++)
	      {
		if (obuf[i] & 0x80)
		{
		  if ((obuf[i] & 0x7f) != '\\'
		      && (0x20 <= (obuf[i] & 0x7f)
			  && (obuf[i] & 0x7f) <= 0x7E))
		  {
		    printf ("%c", obuf[i] & 0x7f);
		  }
		  else
		    printf ("\\%o", obuf[i] & 0x7f);
		}
		else
		  printf ("%c", obuf[i]);
	      }

	    printf ("\"");

	    if (mangle)
	      printf (")");

	    if (showlen || incl_orig)
	      printf ("\n");
	    printf ("\n");

	    olen = 0;
	    ilen = 0;
	    was = 0;
	  }
	}
      }
  }

  fclose (in);
}
